"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const request_promise_1 = __importDefault(require("request-promise"));
const querystring_1 = __importDefault(require("querystring"));
const https_1 = require("https");
const helpers_1 = require("../helpers");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const CFError = require('cf-errors');
const logger = helpers_1.createNewLogger('codefresh:infra:git-providers:gitlab');
const MAX_PER_PAGE = 100;
const MAX_CONCURRENT_REQUESTS = 4; // requests X pages each time
const MAX_RESULTS = 10000;
const statesMap = {
    pending: 'pending',
    running: 'running',
    success: 'success',
    failure: 'failed',
    error: 'failed',
};
const _extractErrorFromResponse = (res) => {
    var _a, _b, _c;
    let err = `status: ${res.statusCode}`;
    if (((_a = res.body) === null || _a === void 0 ? void 0 : _a.error) || ((_b = res.body) === null || _b === void 0 ? void 0 : _b.message)) {
        err += `, err: ${res.body.error || ((_c = res.body) === null || _c === void 0 ? void 0 : _c.message)}`;
    }
    return err;
};
const _toOrderBy = (orderBy) => {
    switch (orderBy) {
        case 'name': return 'name';
        case 'created': return 'created_at';
        case 'pushed': return 'last_activity_at';
        default: return 'last_activity_at';
    }
};
const _encodeProjectPath = (owner, repo) => {
    return encodeURIComponent(`${owner}/${repo}`).replace(/\./g, '%2E');
};
const _cleanEncodedFilePath = (filepath) => {
    const prefix = './';
    const hasPrefix = filepath.indexOf(prefix) === 0;
    const path = hasPrefix ? filepath.slice(prefix.length) : filepath;
    return encodeURIComponent(path);
};
const _toBranch = (rawBranch) => {
    return {
        name: rawBranch.name,
        id: rawBranch.name,
        commit: {
            sha: rawBranch.commit.id,
            commiter_name: rawBranch.commit.author_name,
            message: rawBranch.commit.message,
            url: rawBranch.commit.web_url,
        }
    };
};
const _toUser = (user, email) => {
    return {
        login: user.username,
        email: email || user.email,
        avatar_url: user.avatar_url,
        web_url: user.web_url,
    };
};
const _toRepo = (rawRepo) => {
    const isGroup = lodash_1.get(rawRepo, 'namespace.kind') === 'group'; // is it a group/org repository
    return {
        id: rawRepo.id.toString(),
        provider: 'gitlab',
        name: rawRepo.name,
        full_name: rawRepo.path_with_namespace,
        private: rawRepo.visibility === 'private',
        pushed_at: rawRepo.last_activity_at,
        open_issues: rawRepo.open_issues_count,
        clone_url: rawRepo.http_url_to_repo,
        ssh_url: rawRepo.ssh_url_to_repo,
        owner: {
            login: lodash_1.get(rawRepo, isGroup ? 'namespace.full_path' : 'owner.username', ''),
            avatar_url: lodash_1.get(rawRepo, isGroup ? 'namespace.avatar_url' : 'owner.avatar_url', ''),
            creator: lodash_1.get(rawRepo, isGroup ? 'namespace.name' : 'owner.name', null),
        },
        org: isGroup ? lodash_1.get(rawRepo, 'namespace.name') : null,
        default_branch: rawRepo.default_branch,
        permissions: {
            admin: lodash_1.get(rawRepo, `permissions.${isGroup ? 'group_access' : 'project_access'}.access_level`, 0) >= 40,
        },
        webUrl: rawRepo.web_url || '',
    };
};
const _toWebhook = (owner, repo, rawWebhook) => {
    return {
        id: rawWebhook.id,
        name: String(rawWebhook.id),
        endpoint: rawWebhook.url,
        repo: `${owner}/${repo}`,
        events: Object.keys({
            ...(rawWebhook.push_events && { push_events: true }),
            ...(rawWebhook.tag_push_events && { tag_push_events: true }),
            ...(rawWebhook.merge_requests_events && { merge_requests_events: true }),
            ...(rawWebhook.issues_events && { issues_events: true }),
            ...(rawWebhook.releases_events && { releases_events: true }),
        }),
    };
};
class Gitlab {
    constructor(opts) {
        this.baseUrl = opts.apiURL || 'https://gitlab.com/api/v4/';
        if (!this.baseUrl.endsWith('/')) {
            this.baseUrl = `${this.baseUrl}/`;
        }
        this.timeout = opts.timeout || 10000;
        // works for both oauth2 and private token authentication methods
        // eslint-disable-next-line @typescript-eslint/naming-convention
        this.authenticationHeader = opts.password ? { Authorization: `Bearer ${opts.password}` } : {};
        if (process.env.NODE_TLS_REJECT_UNAUTHORIZED === '0' || opts.insecure) {
            this.agent = new https_1.Agent({ rejectUnauthorized: false });
        }
    }
    async performAPICall(opts) {
        const method = opts.method || 'GET';
        const requestHeaders = lodash_1.merge(opts.noAuth ? {} : this.authenticationHeader, opts.headers || {});
        const requestOptions = {
            method,
            url: `${this.baseUrl}${opts.api}`,
            qs: opts.qs || {},
            headers: requestHeaders,
            json: opts.json,
            timeout: this.timeout,
            form: opts.data,
            resolveWithFullResponse: true,
            agent: this.agent,
            simple: false,
            body: opts.body,
        };
        let requestStr = `${method} ${requestOptions.url}`;
        if (Object.keys(requestOptions.qs).length) {
            requestStr += `?${querystring_1.default.stringify(requestOptions.qs)}`;
        }
        logger.debug(`request: ${requestStr}`);
        return request_promise_1.default(requestOptions)
            .then((res) => {
            logger.debug(`request: ${requestStr} status: ${res.statusCode}`);
            return res;
        });
    }
    async getPaginatedResults(apiFunc, args, limit = MAX_RESULTS, page = 1) {
        const finalResults = [];
        let startingPage = page;
        let shouldStop = false;
        while (!shouldStop) {
            const neededResults = limit - finalResults.length;
            // make up to MAX_PAGES_PER_CYCLE requests in parallel
            const neededPages = Math.min(lodash_1.ceil(neededResults / MAX_PER_PAGE), MAX_CONCURRENT_REQUESTS);
            const pagePromises = lodash_1.range(startingPage, startingPage + neededPages).map((pageNumber) => apiFunc(lodash_1.merge(args, {
                qs: {
                    page: pageNumber,
                    per_page: MAX_PER_PAGE,
                }
            })));
            // harvest results
            const pagesResults = await Promise.all(pagePromises);
            for (const res of pagesResults) {
                if (res.statusCode >= 400) {
                    return res;
                }
                const currentNeededResults = limit - finalResults.length;
                const data = res.body;
                finalResults.push(...data.slice(0, currentNeededResults));
                if (limit === finalResults.length) {
                    shouldStop = true;
                    break; // got all the results we need, no reason to process other requests
                }
                if (data.length < MAX_PER_PAGE) {
                    shouldStop = true;
                    break; // last page, no reason to process other requests
                }
            }
            // next batch
            startingPage += neededPages;
        }
        return {
            data: finalResults,
            statusCode: 200,
        };
    }
    getName() {
        return 'gitlab';
    }
    async fetchRawFile(opt) {
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(opt.owner, opt.repo)}/repository/files/${_cleanEncodedFilePath(opt.path)}`,
            qs: { ref: opt.ref },
            json: true,
        });
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to retrieve file: ${JSON.stringify(opt)}, status code: ${res.statusCode}`);
        }
        const base64Content = res.body.content.replace(/\n/gi, '');
        return Buffer.from(base64Content, 'base64').toString();
    }
    async getBranch(opt) {
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(opt.owner, opt.repo)}/repository/branches/${opt.branch}`,
            json: true
        });
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opt)}, status code: ${res.statusCode}`);
        }
        return _toBranch(res.body);
    }
    async getRepository(opt) {
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(opt.owner, opt.repo)}`,
            json: true,
        });
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to get repository: ${JSON.stringify(opt)}, ${_extractErrorFromResponse(res)}`);
        }
        return _toRepo(res.body);
    }
    async listBranches(opt) {
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(opt.owner, opt.repo)}/repository/branches`,
            json: true,
        });
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to list repository branches: ${JSON.stringify(opt)}, ${_extractErrorFromResponse(res)}`);
        }
        return res.body.map(_toBranch);
    }
    async listRepositoriesForOwner(opt) {
        throw new Error('Method not implemented.');
    }
    async listWebhooks(opt) {
        const { owner, repo } = opt;
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(owner, repo)}/hooks`,
            json: true,
        });
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to list repository webhooks: ${JSON.stringify(opt)}, ${_extractErrorFromResponse(res)}`);
        }
        return res.body.map(_toWebhook.bind(undefined, owner, repo));
    }
    async createRepositoryWebhook(opt) {
        const { owner, repo, endpoint, secret } = opt;
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(owner, repo)}/hooks`,
            method: 'POST',
            json: true,
            body: {
                url: endpoint,
                token: secret,
                push_events: true,
                merge_requests_events: true,
                tag_push_events: true,
                description: 'Codefresh',
            }
        });
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to create repository webhook: ${JSON.stringify(opt)}, ${_extractErrorFromResponse(res)}`);
        }
        return _toWebhook(owner, repo, res.body);
    }
    async deleteRepositoryWebhook(opt) {
        const { owner, repo, hookId } = opt;
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(owner, repo)}/hooks/${hookId}`,
            method: 'DELETE',
        });
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to delete repository webhook: ${JSON.stringify(opt)}, ${_extractErrorFromResponse(res)}`);
        }
    }
    async listRepositoriesWithAffiliation(opt) {
        const { affiliation, sort, direction, limit, page } = opt;
        const [err, res] = await helpers_1.to(this.getPaginatedResults(this.performAPICall.bind(this), {
            api: 'projects/',
            json: true,
            qs: {
                membership: true,
                order_by: _toOrderBy(sort),
                simple: false,
                ...(direction && { sort: direction }),
                ...(affiliation === 'owner' && { owned: true })
            },
        }, limit, page));
        if (err) {
            throw new CFError(`Failed to list repositories: ${JSON.stringify(opt)}, err: ${err}`);
        }
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to list repositories: ${JSON.stringify(opt)}, ${_extractErrorFromResponse(res)}`);
        }
        return res.data.map(_toRepo);
    }
    async listOrganizations(opt) {
        throw new Error('Method not implemented.');
    }
    async listRepositoriesForOrganization(opt) {
        throw new Error('Method not implemented.');
    }
    async createCommitStatus(opt) {
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(opt.owner, opt.repo)}/statuses/${opt.sha}`,
            method: 'POST',
            json: true,
            body: {
                context: opt.context,
                description: opt.description,
                target_url: opt.targetUrl,
                state: statesMap[opt.state],
            }
        });
        const error = _extractErrorFromResponse(res);
        if (res.statusCode === 400 && error.includes('running') && error.includes('Cannot transition status')) {
            /* for fixing pipelines running forever,
            avoid retry mechanism in case of error that cause by change commit status from running to running
            https://codefresh-io.atlassian.net/browse/CR-5024?atlOrigin=eyJpIjoiNmNkODA2NzdhNmVlNGVhNWFiYjljZjE3OTc3ZmNlZDkiLCJwIjoiaiJ9 */
            return;
        }
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to create commit status: ${JSON.stringify(opt)}, ${error}`);
        }
    }
    async getUser(opt) {
        const user = (opt === null || opt === void 0 ? void 0 : opt.username) ? opt.username : 'authenticated user';
        const [err, res] = await helpers_1.to(this.performAPICall({
            api: (opt === null || opt === void 0 ? void 0 : opt.username) ? 'users' : 'user',
            json: true,
            qs: (opt === null || opt === void 0 ? void 0 : opt.username) ? { username: opt.username } : {}
        }));
        if (err) {
            throw new CFError({
                message: `Failed to get ${user}: , status code: ${err.statusCode}`,
                cause: err,
            });
        }
        else if (res.statusCode >= 400) {
            throw new CFError(`Failed to get user: ${user}, status code: ${res.statusCode}`);
        }
        return _toUser(res.body);
    }
    async getUserByEmail(email) {
        const [err, res] = await helpers_1.to(this.performAPICall({
            api: 'users',
            json: true,
            qs: { search: email }
        }));
        if (err) {
            throw new CFError({
                message: `Failed to get user with email ${email}: , status code: ${err.statusCode}`,
                cause: err,
            });
        }
        else if (res.statusCode >= 400) {
            throw new CFError(`Failed to get user with email: ${email}, status code: ${res.statusCode}`);
        }
        if (res.body.length === 0) {
            throw new CFError(`No user with email: ${email} was found on ${this.getName()}`);
        }
        return _toUser(res.body[0], email);
    }
    async getPullRequestFiles(opt) {
        throw new Error('Method not implemented.');
    }
    async getPullRequest(opt) {
        throw new Error('Method not implemented.');
    }
    async getRepositoryPermissions(opt) {
        throw new Error('Method not implemented.');
    }
    async assertApiScopes(opt) {
        //TODO : Impl validation
    }
}
exports.default = Gitlab;
//# sourceMappingURL=gitlab.js.map